
# q-expansion of newform 108.4.a.d, downloaded from the LMFDB on 14 May 2024.

# We generate the q-expansion using the Hecke eigenvalues a_p at the primes.
# Each a_p is given as a linear combination
# of the following basis for the coefficient ring.

def make_data():

    from sage.all import prod, floor, prime_powers, gcd, QQ, primes_first_n, next_prime, RR

    def discrete_log(elts, gens, mod):
        # algorithm 2.2, page 16 of https://arxiv.org/abs/0903.2785
        def table_gens(gens, mod):
            T = [1]
            n = len(gens)
            r = [None]*n
            s = [None]*n
            for i in range(n):
                beta = gens[i]
                r[i] = 1
                N = len(T)
                while beta not in T:
                    for Tj in T[:N]:
                        T.append((beta*Tj) % mod)
                    beta = (beta*gens[i]) % mod
                    r[i] += 1
                s[i] = T.index(beta)
            return T, r, s
        T, r, s = table_gens(gens, mod)
        n = len(gens)
        N = [ prod(r[:j]) for j in range(n) ]
        Z = lambda s: [ (floor(s/N[j]) % r[j]) for j in range(n)]
        return [Z(T.index(elt % mod)) for elt in elts]
    def extend_multiplicatively(an):
        for pp in prime_powers(len(an)-1):
            for k in range(1, (len(an) - 1)//pp + 1):
                if gcd(k, pp) == 1:
                    an[pp*k] = an[pp]*an[k]
    from sage.all import PolynomialRing, NumberField
    R = PolynomialRing(QQ, "x")
    f = R(poly_data)
    K = NumberField(f, "a")
    betas = [K.gens()[0]**i for i in range(len(poly_data))]
    convert_elt_to_field = lambda elt: sum(c*beta for c, beta in zip(elt, betas))
    # convert aps to K elements
    primes = primes_first_n(len(aps_data))
    good_primes = [p for p in primes if not p.divides(level)]
    aps = map(convert_elt_to_field, aps_data)
    if not hecke_ring_character_values:
        # trivial character
        char_values = dict(zip(good_primes, [1]*len(good_primes)))
    else:
        gens = [elt[0] for elt in hecke_ring_character_values]
        gens_values = [convert_elt_to_field(elt[1]) for elt in hecke_ring_character_values]
        char_values = dict([(
            p,prod(g**k for g, k in zip(gens_values, elt)))
            for p, elt in zip(good_primes, discrete_log(good_primes, gens, level))
            ])
    an_list_bound = next_prime(primes[-1])
    an = [0]*an_list_bound
    an[1] = 1
    
    from sage.all import PowerSeriesRing
    PS = PowerSeriesRing(K, "q")
    for p, ap in zip(primes, aps):
        if p.divides(level):
            euler_factor = [1, -ap]
        else:
            euler_factor = [1, -ap, p**(weight - 1) * char_values[p]]
        k = RR(an_list_bound).log(p).floor() + 1
        foo = (1/PS(euler_factor)).padded_list(k)
        for i in range(1, k):
            an[p**i] = foo[i]
    extend_multiplicatively(an)
    return PS(an)
level = 108
weight = 4
poly_data = [0, 1]

# The basis for the coefficient ring is just the power basis
# in the root of the defining polynomial above.
hecke_ring_character_values = None
aps_data = [[0], [0], [9], [-1], [63], [-28], [72], [98], [126], [-126], [-259], [386], [-450], [-34], [-54], [-693], [180], [-280], [-586], [504], [161], [440], [999], [882], [-721], [-441], [-532], [819], [-1294], [1134], [-1807], [-2205], [1386], [476], [-1575], [449], [1820], [-1828], [-810], [-1323], [-315], [-2800], [-3276], [3221], [3339], [3689], [-6022], [-952], [5292], [2198], [5166], [3402], [1862], [-5472], [5292], [1638], [-1206], [4319], [-2248], [-2016], [-2338], [-8514], [6104], [-4338], [-8155], [-4977], [5678], [2906], [6993], [7910], [2466], [7182], [-11431], [-6616], [-9820], [1440], [5985], [-11284], [-7308], [6335], [6372], [3320], [-2898], [-4291], [-8323], [-3780], [12474], [16679], [-17271], [17387], [3087], [-5238], [-4384], [-3843], [12566], [-5238], [-6309], [-5868], [6776], [-9388], [15056], [16569], [14553], [4284], [692], [4886], [-11025], [-14562], [11466], [10955], [1232], [25622], [21042], [6524], [20663], [-882], [-7252], [21168], [14301], [15057], [-4690], [-5203], [6174], [756], [17948], [34083], [-5284], [-24516], [-12481], [-3094], [-376], [29484], [11681], [890], [-32832], [3185], [-21222], [12320], [-11907], [-14994], [-38878], [3906], [-10207], [8064], [-10486], [-33696], [-15778], [-26136], [23912], [17892], [44162], [-8820], [-4654], [33084], [47258], [36666], [-2995], [-13284], [-2989], [46485], [11529], [37044], [12455], [-20169], [-39942], [54234], [-26137], [-33460]]
